@@ -1,4 +1,5 @@
 const Applet = imports.ui.applet;
+const Lang = imports.lang;
 const Mainloop = imports.mainloop;
 const Gio = imports.gi.Gio;
 const Interfaces = imports.misc.interfaces;
@@ -167,13 +168,14 @@ class VolumeSlider extends PopupMenu.PopupSliderMenuItem {
         }
 
         this._slider.queue_repaint();
+        this.tooltip.show();
         this.emit('value-changed', this._value);
     }
 
     _onKeyPressEvent(actor, event) {
         let key = event.get_key_symbol();
-        if (key == Clutter.KEY_Right || key == Clutter.KEY_Left) {
-            let delta = key == Clutter.KEY_Right ? VOLUME_ADJUSTMENT_STEP : -VOLUME_ADJUSTMENT_STEP;
+        if (key === Clutter.KEY_Right || key === Clutter.KEY_Left) {
+            let delta = key === Clutter.KEY_Right ? VOLUME_ADJUSTMENT_STEP : -VOLUME_ADJUSTMENT_STEP;
             this._value = Math.max(0, Math.min(this._value + delta/this.applet._volumeMax*this.applet._volumeNorm, 1));
             this._slider.queue_repaint();
             this.emit('value-changed', this._value);
@@ -208,6 +210,8 @@ class VolumeSlider extends PopupMenu.PopupSliderMenuItem {
         let percentage = Math.round(visible_value * 100) + "%";
 
         this.tooltip.set_text(this.tooltipText + percentage);
+        if (this._dragging)
+            this.tooltip.show();
         let iconName = this._volumeToIcon(value);
         if (this.app_icon == null) {
             this.icon.icon_name = iconName;
@@ -421,7 +425,7 @@ class StreamMenuSection extends PopupMenu.PopupMenuSection {
         }
 
         // Trim stream name
-        if(name.length > 16) {
+        if(name.length > 20) {
             name = name.substring(0, 16) + "... ";
         }
 
@@ -496,34 +500,20 @@ class Player extends PopupMenu.PopupMenuSection {
         mainBox.addActor(this.vertBox, { expand: false });
 
         // Player info
-        let playerBox = new St.BoxLayout();
+        this._playerBox = new St.BoxLayout();
         this.playerIcon = new St.Icon({icon_type: St.IconType.SYMBOLIC, style_class: "popup-menu-icon"});
         this.playerLabel = new St.Label({ y_expand: true, y_align: Clutter.ActorAlign.CENTER, x_expand: true });
-        playerBox.add_actor(this.playerIcon);
-        playerBox.add_actor(this.playerLabel);
+        this._playerBox.add_actor(this.playerIcon);
+        this._playerBox.add_actor(this.playerLabel);
 
         if (this._mediaServer.CanRaise) {
-            let btn = new ControlButton("go-up", _("Open Player"), () => {
-                if (this._name.toLowerCase() === "spotify") {
-                    // Spotify isn't able to raise via Dbus once its main UI is closed
-                    Util.spawn(['spotify']);
-                }
-                else {
-                    this._mediaServer.RaiseRemote();
-                }
-                this._applet.menu.close();
-            }, true);
-            playerBox.add_actor(btn.actor);
+            this._showCanRaise();
         }
         if (this._mediaServer.CanQuit) {
-            let btn = new ControlButton("window-close", _("Quit Player"), () => {
-                this._mediaServer.QuitRemote();
-                this._applet.menu.close();
-            }, true);
-            playerBox.add_actor(btn.actor);
+            this._showCanQuit();
         }
 
-        this.vertBox.add_actor(playerBox);
+        this.vertBox.add_actor(this._playerBox);
 
         // Cover Box (art + track info)
         this._trackCover = new St.Bin({x_align: St.Align.MIDDLE});
@@ -579,20 +569,12 @@ class Player extends PopupMenu.PopupMenuSection {
         this.controls.add_actor(this._stopButton.getActor());
         this.controls.add_actor(this._nextButton.getActor());
         trackControls.set_child(this.controls);
-        if (this._mediaServerPlayer.LoopStatus) {
-            this._loopButton = new ControlButton("media-playlist-consecutive", _("Consecutive Playing"), () => this._toggleLoopStatus());
-            this._loopButton.actor.visible = this._applet.extendedPlayerControl;
-            this.controls.add_actor(this._loopButton.getActor());
 
-            this._setLoopStatus(this._mediaServerPlayer.LoopStatus);
-        }
-        if (this._mediaServerPlayer.Shuffle !== undefined) {
-            this._shuffleButton = new ControlButton("media-playlist-shuffle", _("No Shuffle"), () => this._toggleShuffle());
-            this._shuffleButton.actor.visible = this._applet.extendedPlayerControl;
-            this.controls.add_actor(this._shuffleButton.getActor());
+        this._loopButton = new ControlButton("media-playlist-consecutive", _("Consecutive Playing"), () => this._toggleLoopStatus());
+        this.controls.add_actor(this._loopButton.getActor());
 
-            this._setShuffle(this._mediaServerPlayer.Shuffle);
-        }
+        this._shuffleButton = new ControlButton("media-playlist-shuffle", _("No Shuffle"), () => this._toggleShuffle());
+        this.controls.add_actor(this._shuffleButton.getActor());
 
         // Position slider
         this._seeker = new Seeker(this._mediaServerPlayer, this._prop, this._name.toLowerCase());
@@ -614,13 +596,49 @@ class Player extends PopupMenu.PopupMenuSection {
                 this._setLoopStatus(props.LoopStatus.unpack());
             if (props.Shuffle)
                 this._setShuffle(props.Shuffle.unpack());
+            if (props.Identity) {
+                this._name = props.Identity.unpack();
+                this._applet._updatePlayerMenuItems();
+            }
+            if (props.CanRaise) {
+                this._showCanRaise();
+            }
+            if (props.CanQuit) {
+                this._showCanQuit();
+            }
+            if (props.DesktopEntry) {
+                this._applet.passDesktopEntry(props.DesktopEntry.unpack());
+            }
         });
 
-        //get the desktop entry and pass it to the applet
-        this._prop.GetRemote(MEDIA_PLAYER_2_NAME, "DesktopEntry", (result, error) => {
-            if (!error)
-                this._applet.passDesktopEntry(result[0].unpack());
-        });
+        this._setLoopStatus(this._mediaServerPlayer.LoopStatus);
+        this._setShuffle(this._mediaServerPlayer.Shuffle);
+
+        if (this._mediaServer.DesktopEntry) {
+            this._applet.passDesktopEntry(this._mediaServer.DesktopEntry);
+        }
+    }
+
+    _showCanRaise() {
+        let btn = new ControlButton("go-up", _("Open Player"), () => {
+            if (this._name.toLowerCase() === "spotify") {
+                // Spotify isn't able to raise via Dbus once its main UI is closed
+                Util.spawn(['spotify']);
+            }
+            else {
+                this._mediaServer.RaiseRemote();
+            }
+            this._applet.menu.close();
+        }, true);
+        this._playerBox.add_actor(btn.actor);
+    }
+
+    _showCanQuit() {
+        let btn = new ControlButton("window-close", _("Quit Player"), () => {
+            this._mediaServer.QuitRemote();
+            this._applet.menu.close();
+        }, true);
+        this._playerBox.add_actor(btn.actor);
     }
 
     _setName(status) {
@@ -691,9 +709,8 @@ class Player extends PopupMenu.PopupMenuSection {
         let change = false;
         if (metadata["mpris:artUrl"]) {
             let artUrl = metadata["mpris:artUrl"].unpack();
-            if ( this._name === "spotify" ) {
-                artUrl = artUrl.replace("/thumb/", "/300/"); // Spotify 0.9.x
-                artUrl = artUrl.replace("/image/", "/300/"); // Spotify 0.27.x
+            if ( this._name.toLowerCase() === "spotify" ) {
+                artUrl = artUrl.replace("open.spotify.com", "i.scdn.co");
             }
             if (this._trackCoverFile != artUrl) {
                 this._trackCoverFile = artUrl;
@@ -767,6 +784,8 @@ class Player extends PopupMenu.PopupMenuSection {
     }
 
     _setLoopStatus(status) {
+        this._loopButton.actor.visible = this._applet.extendedPlayerControl && this._mediaServerPlayer.LoopStatus;
+
         if(status === "None")
             this._loopButton.setData("media-playlist-consecutive-symbolic", _("Consecutive Playing"));
         else if(status === "Track")
@@ -782,6 +801,8 @@ class Player extends PopupMenu.PopupMenuSection {
     }
 
     _setShuffle(status) {
+        this._shuffleButton.actor.visible = this._applet.extendedPlayerControl && this._mediaServerPlayer.Shuffle;
+
         this._shuffleButton.setData("media-playlist-shuffle", status? _("Shuffle") : _("No Shuffle"));
         this._shuffleButton.setActive(status);
     }
@@ -811,8 +832,8 @@ class Player extends PopupMenu.PopupMenuSection {
     }
 
     onSettingsChanged() {
-        this._loopButton.actor.visible = this._applet.extendedPlayerControl;
-        this._shuffleButton.actor.visible = this._applet.extendedPlayerControl;
+        this._loopButton.actor.visible = this._applet.extendedPlayerControl && this._mediaServerPlayer.LoopStatus;
+        this._shuffleButton.actor.visible = this._applet.extendedPlayerControl && this._mediaServerPlayer.Shuffle;
     }
 
     destroy() {
@@ -834,11 +855,12 @@ class MediaPlayerLauncher extends PopupMenu.PopupBaseMenuItem {
         this.addActor(this.label);
         this._icon = app.create_icon_texture(ICON_SIZE);
         this.addActor(this._icon, { expand: false });
+        this.connect("activate", (event) => this._onActivate(event));
     }
 
-    //note: shadows base method and prevents "activate" emission
-    activate(event, keepMenu) {
-        this._app.activate_full(-1, event.get_time());
+    _onActivate(event) {
+        let _time = event.time;
+        this._app.activate_full(-1, _time);
     }
 }
 
@@ -869,10 +891,13 @@ class CinnamonSoundApplet extends Applet.TextIconApplet {
 
         this.settings.bind("_knownPlayers", "_knownPlayers");
         if (this.hideSystray) this.registerSystrayIcons();
+        
+        this.settings.bind("keyOpen", "keyOpen", this._setKeybinding);
 
         this.menuManager = new PopupMenu.PopupMenuManager(this);
         this.menu = new Applet.AppletPopupMenu(this, orientation);
         this.menuManager.addMenu(this.menu);
+        this._setKeybinding();
 
         this.set_applet_icon_symbolic_name('audio-x-generic');
 
@@ -996,6 +1021,10 @@ class CinnamonSoundApplet extends Applet.TextIconApplet {
 
         this._sound_settings.connect("changed::" + MAXIMUM_VOLUME_KEY, () => this._on_sound_settings_change());
     }
+    
+    _setKeybinding() {
+        Main.keybindingManager.addHotKey("sound-open-" + this.instance_id, this.keyOpen, Lang.bind(this, this._openMenu));
+    }
 
     _on_sound_settings_change () {
         this._volumeMax = this._sound_settings.get_int(MAXIMUM_VOLUME_KEY) / 100 * this._control.get_vol_max_norm();
@@ -1018,6 +1047,7 @@ class CinnamonSoundApplet extends Applet.TextIconApplet {
     }
 
     on_applet_removed_from_panel () {
+        Main.keybindingManager.removeHotKey("sound-open-" + this.instance_id);
         if (this.hideSystray)
             this.unregisterSystrayIcons();
         if (this._iconTimeoutId) {
@@ -1031,6 +1061,10 @@ class CinnamonSoundApplet extends Applet.TextIconApplet {
     }
 
     on_applet_clicked(event) {
+        this._openMenu();
+    }
+    
+    _openMenu() {
         this.menu.toggle();
     }
 
@@ -1095,6 +1129,8 @@ class CinnamonSoundApplet extends Applet.TextIconApplet {
             }
         }
 
+        this._applet_tooltip.show();
+
         this._notifyVolumeChange(this._output);
     }
 
@@ -1309,15 +1345,15 @@ class CinnamonSoundApplet extends Applet.TextIconApplet {
     }
 
     _showFixedElements() {
-        // The list to use when switching between active players
-        this._chooseActivePlayerItem = new PopupMenu.PopupSubMenuMenuItem(_("Choose player controls"));
-        this._chooseActivePlayerItem.actor.hide();
-        this.menu.addMenuItem(this._chooseActivePlayerItem);
-
         // The launch player list
         this._launchPlayerItem = new PopupMenu.PopupSubMenuMenuItem(_("Launch player"));
         this.menu.addMenuItem(this._launchPlayerItem);
         this._updateLaunchPlayer();
+        
+        // The list to use when switching between active players
+        this._chooseActivePlayerItem = new PopupMenu.PopupSubMenuMenuItem(_("Choose player controls"));
+        this._chooseActivePlayerItem.actor.hide();
+        this.menu.addMenuItem(this._chooseActivePlayerItem);
 
         //between these two separators will be the player MenuSection (position 3)
         this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
@@ -1349,14 +1385,16 @@ class CinnamonSoundApplet extends Applet.TextIconApplet {
                 let menuItem = new MediaPlayerLauncher(playerApp, this._launchPlayerItem.menu);
                 this._launchPlayerItem.menu.addMenuItem(menuItem);
             }
-        } else {
+        }
+
+        if (!this.playerControl || !availablePlayers.length) {
             this._launchPlayerItem.actor.hide();
         }
     }
 
     _updatePlayerMenuItems() {
         if (this.playerControl && this._activePlayer) {
-            this._launchPlayerItem.actor.hide();
+            this._launchPlayerItem.actor.show();
             this._chooseActivePlayerItem.actor.show();
 
             // Show a dot on the active player in the switching menu
@@ -1376,6 +1414,7 @@ class CinnamonSoundApplet extends Applet.TextIconApplet {
                 this._launchPlayerItem.actor.show();
             } else {
                 this._launchPlayerItem.actor.hide();
+                this._chooseActivePlayerItem.actor.hide();
             }
         }
     }
@@ -1387,7 +1426,7 @@ class CinnamonSoundApplet extends Applet.TextIconApplet {
         this._activePlayer = player;
         if (this.playerControl && this._activePlayer != null) {
             let menuItem = this._players[player];
-            this.menu.addMenuItem(menuItem, 1);
+            this.menu.addMenuItem(menuItem, 2);
         }
 
         this._updatePlayerMenuItems();
