@@ -1,23 +1,19 @@
 #!/usr/bin/python3
 
-import sys
 import os
 import re
-import json
 import html
 import subprocess
 import gettext
 from html.parser import HTMLParser
 import html.entities as entities
 
-import dbus
 import gi
 gi.require_version('Gtk', '3.0')
-from gi.repository import Gio, Gtk, GObject, Gdk, GdkPixbuf, Pango, GLib
+from gi.repository import Gio, Gtk, GdkPixbuf, GLib
 
-from xapp.SettingsWidgets import SettingsStack, SettingsPage, SettingsWidget, SettingsLabel
-from SettingsWidgets import SidePage
-from Spices import Spice_Harvester, ThreadedTaskManager
+from xapp.SettingsWidgets import SettingsPage, SettingsWidget, SettingsLabel
+from Spices import ThreadedTaskManager
 
 home = os.path.expanduser('~')
 
@@ -34,6 +30,9 @@ ROW_SIZE = 32
 UNSAFE_ITEMS = ['spawn_sync', 'spawn_command_line_sync', 'GTop', 'get_file_contents_utf8_sync']
 
 curr_ver = subprocess.check_output(['cinnamon', '--version']).decode("utf-8").splitlines()[0].split(' ')[1]
+curr_ver_elements = curr_ver.split(".")
+curr_ver_major = int(curr_ver_elements[0])
+curr_ver_minor = int(curr_ver_elements[1])
 
 def find_extension_subdir(directory):
     largest = ['0']
@@ -95,7 +94,6 @@ def show_prompt(msg, window=None):
                                destroy_with_parent = True,
                                message_type = Gtk.MessageType.QUESTION,
                                buttons = Gtk.ButtonsType.YES_NO)
-    dialog.set_default_size(400, 200)
     esc = html.escape(msg)
     dialog.set_markup(esc)
     dialog.show_all()
@@ -103,6 +101,17 @@ def show_prompt(msg, window=None):
     dialog.destroy()
     return response == Gtk.ResponseType.YES
 
+def show_message(msg, window=None):
+    dialog = Gtk.MessageDialog(transient_for = window,
+                               destroy_with_parent = True,
+                               message_type = Gtk.MessageType.ERROR,
+                               buttons = Gtk.ButtonsType.OK)
+    esc = html.escape(msg)
+    dialog.set_markup(esc)
+    dialog.show_all()
+    dialog.run()
+    dialog.destroy()
+
 background_work_queue = ThreadedTaskManager(5)
 
 
@@ -163,8 +172,8 @@ class ManageSpicesRow(Gtk.ListBoxRow):
         except (KeyError, ValueError):
             last_edited = -1
 
-        if 'multiversion' in self.metadata and self.metadata['multiversion']:
-            self.metadata['path'] = find_extension_subdir(self.metadata['path'])
+        # Check for the right version subdir (if the spice is multi-versioned, it won't necessarily be in its root directory)
+        self.metadata['path'] = find_extension_subdir(self.metadata['path'])
 
         # "hide-configuration": true in metadata trumps all
         # otherwise we check for "external-configuration-app" in metadata and settings-schema.json in settings
@@ -257,11 +266,30 @@ class ManageSpicesRow(Gtk.ListBoxRow):
         if self.writable:
             self.scan_extension_for_danger(self.metadata['path'])
 
-        self.version_supported = False
+        self.version_supported = self.is_compatible_with_cinnamon_version()
+
+    def is_compatible_with_cinnamon_version(self):
         try:
-            self.version_supported = curr_ver in self.metadata['cinnamon-version'] or curr_ver.rsplit('.', 1)[0] in self.metadata['cinnamon-version']
-        except (KeyError, ValueError):
-            self.version_supported = True # Don't check version if not specified.
+            # Treat "cinnamon-version" as a list of minimum required versions
+            # if any version in there is lower than our Cinnamon version, then the spice is compatible.
+            for version in self.metadata['cinnamon-version']:
+                elements = version.split(".")
+                major = int(elements[0])
+                minor = int(elements[1])
+                if curr_ver_major > major or (curr_ver_major == major and curr_ver_minor >= minor):
+                    # The version is OK, check that we can find the right .js file in the appropriate subdir
+                    path = os.path.join(self.metadata['path'], self.extension_type + ".js")
+                    if os.path.exists(path):
+                        return True
+                    else:
+                        print ("The %s %s is not properly structured. Path not found: '%s'" % (self.uuid, self.extension_type, path))
+                        return False
+                    return True
+            print ("The %s %s is not compatible with this version of Cinnamon." % (self.uuid, self.extension_type))
+            return False
+        except:
+            # If cinnamon-version is not specified or if the version check goes wrong, assume compatibility
+            return True
 
     def set_can_config(self, *args):
         if not self.has_config:
@@ -510,10 +538,8 @@ class ManageSpicesPage(SettingsPage):
 
     def enable_extension(self, uuid, name, version_check = True):
         if not version_check:
-            if not show_prompt(_("Extension %s is not compatible with current version of cinnamon. Using it may break your system. Load anyway?") % uuid, self.window):
-                return
-            else:
-                uuid = '!' + uuid
+            show_message(_("Extension %s is not compatible with your version of Cinnamon.") % uuid, self.window)
+            return
 
         self.enable(uuid)
 
@@ -546,9 +572,10 @@ class ManageSpicesPage(SettingsPage):
         elif self.collection_type == 'extension':
             msg = _("This will disable all active extensions. Are you sure you want to do this?")
         if show_prompt(msg, self.window):
+            sett = Gio.Settings.new('org.cinnamon')
             if self.collection_type != 'extension':
-                os.system(('gsettings reset org.cinnamon next-%s-id') % (self.collection_type))
-            os.system(('gsettings reset org.cinnamon enabled-%ss') % (self.collection_type))
+                sett.reset('next-%s-id' % self.collection_type)
+            sett.reset('enabled-%ss' % self.collection_type)
 
     def about(self, *args):
         row = self.list_box.get_selected_row()
@@ -961,4 +988,3 @@ class DownloadSpicesPage(SettingsPage):
 
         infobar.set_revealed(False)
         self.search_entry.grab_focus()
-
