@@ -10,11 +10,11 @@ try:
     import html
     import subprocess
     import threading
-    import time
     import dbus
     from PIL import Image
     import datetime
     import proxygsettings
+    import time
 except Exception as detail:
     print(detail)
     sys.exit(1)
@@ -357,6 +357,9 @@ class Spice_Harvester(GObject.Object):
             self._directory_changed()
 
     def _download(self, out_file, url, binary=True):
+        timestamp = round(time.time())
+        url = "%s?time=%d" % (url, timestamp)
+        print("Downloading from %s" % url)
         try:
             open_args = 'wb' if binary else 'w'
             with open(out_file, open_args) as outfd:
@@ -389,7 +392,8 @@ class Spice_Harvester(GObject.Object):
             else:
                 connection = HTTPSConnection(host, timeout=15)
             headers = { "Accept-Encoding": "identity", "Host": host, "User-Agent": "Python/3" }
-            connection.request("GET", parsed_url.path, headers=headers)
+            full_path = "%s?%s" % (parsed_url.path, parsed_url.query)
+            connection.request("GET", full_path, headers=headers)
             urlobj = connection.getresponse()
             assert urlobj.getcode() == 200
 
