@@ -1,6 +1,7 @@
 // -*- mode: js; js-indent-level: 4; indent-tabs-mode: nil -*-
 
 const Clutter = imports.gi.Clutter;
+const GLib = imports.gi.GLib;
 const Gtk = imports.gi.Gtk;
 const St = imports.gi.St;
 const Lang = imports.lang;
@@ -8,54 +9,46 @@ const Cinnamon = imports.gi.Cinnamon;
 const Signals = imports.signals;
 const Tweener = imports.ui.tweener;
 const Main = imports.ui.main;
-const Mainloop = imports.mainloop;
 
 const Params = imports.misc.params;
 
+var DND_ANIMATION_TIME = 0.2;
 // Time to scale down to maxDragActorSize
-const SCALE_ANIMATION_TIME = 0.25;
+var SCALE_ANIMATION_TIME = 0.25;
 // Time to animate to original position on cancel
-const SNAP_BACK_ANIMATION_TIME = 0.25;
+var SNAP_BACK_ANIMATION_TIME = 0.25;
 // Time to animate to original position on success
-const REVERT_ANIMATION_TIME = 0.75;
+var REVERT_ANIMATION_TIME = 0.75;
 
-// Time to wait patiently for events while dragging
-const DRAG_EVENT_TIMEOUT_MS = 5000;
-
-const DragMotionResult = {
-    NO_DROP:   0,
-    COPY_DROP: 1,
-    MOVE_DROP: 2,
-    CONTINUE:  3
+var DragMotionResult = {
+    NO_DROP:       0,
+    COPY_DROP:     1,
+    MOVE_DROP:     2,
+    POINTING_DROP: 3,
+    CONTINUE:      4
 };
 
-const DRAG_CURSOR_MAP = {
+var DRAG_CURSOR_MAP = {
     0: Cinnamon.Cursor.DND_UNSUPPORTED_TARGET,
     1: Cinnamon.Cursor.DND_COPY,
-    2: Cinnamon.Cursor.DND_MOVE
+    2: Cinnamon.Cursor.DND_MOVE,
+    3: Cinnamon.Cursor.POINTING_HAND
 };
 
-const DragDropResult = {
+var DragDropResult = {
     FAILURE:  0,
     SUCCESS:  1,
     CONTINUE: 2
 };
 
-const DND_ANIMATION_TIME = 0.2;
-
-let eventHandlerActor = null;
-let currentDraggable = null;
-let dragMonitors = [];
-
-function isDragging() {
-    return currentDraggable != null;
-}
+var eventHandlerActor = null;
+var currentDraggable = null;
+var dragMonitors = [];
+var targetMonitors = [];
 
 function _getEventHandlerActor() {
     if (!eventHandlerActor) {
-        eventHandlerActor = new Clutter.Rectangle();
-        eventHandlerActor.width = 0;
-        eventHandlerActor.height = 0;
+        eventHandlerActor = new Clutter.Actor({ width: 0, height: 0 });
         Main.uiGroup.add_actor(eventHandlerActor);
         // We connect to 'event' rather than 'captured-event' because the capturing phase doesn't happen
         // when you've grabbed the pointer.
@@ -79,34 +72,42 @@ function removeDragMonitor(monitor) {
         }
 }
 
-function _Draggable(actor, params, target) {
-    this._init(actor, params, target);
+function isDragging() {
+    return currentDraggable != null;
 }
 
-_Draggable.prototype = {
+var _Draggable = new Lang.Class({
+    Name: 'Draggable',
+
     _init : function(actor, params, target) {
-        
-        this.inhibit = false; // Use the inhibit flag to temporarily disable an object from being draggable
-        
         params = Params.parse(params, { manualMode: false,
                                         restoreOnSuccess: false,
+                                        overrideX: undefined,
+                                        overrideY: undefined,
                                         dragActorMaxSize: undefined,
                                         dragActorOpacity: undefined });
 
+        this.inhibit = false; // Use the inhibit flag to temporarily disable an object from being draggable
+
         this.actor = actor;
-        this.target = target;
+        this.target = null;
+        this.recentDropTarget = null;
+
+        if (target) {
+            this.target = target;
+        }
+
+        this.buttonPressEventId = 0;
+        this.destroyEventId = 0;
 
         if (!params.manualMode)
-            this.actor.connect('button-press-event',
-                               Lang.bind(this, this._onButtonPress));
+            this.buttonPressEventId = this.actor.connect('button-press-event',
+                                                    Lang.bind(this, this._onButtonPress));
 
-        this.actor.connect('destroy', Lang.bind(this, function() {
+        this.destroyEventId = this.actor.connect('destroy', Lang.bind(this, function() {
             this._actorDestroyed = true;
-            // If the drag actor is destroyed and we were going to fix
-            // up its hover state, fix up the parent hover state instead
-            if (this.actor == this._firstLeaveActor)
-                this._firstLeaveActor = this._dragOrigParent;
-            if (this._dragInProgress)
+
+            if (this._dragInProgress && this._dragCancellable)
                 this._cancelDrag(global.get_current_time());
             this.disconnectAll();
         }));
@@ -115,54 +116,37 @@ _Draggable.prototype = {
         this._restoreOnSuccess = params.restoreOnSuccess;
         this._dragActorMaxSize = params.dragActorMaxSize;
         this._dragActorOpacity = params.dragActorOpacity;
+        this._overrideX = params.overrideX;
+        this._overrideY = params.overrideY;
 
         this._buttonDown = false; // The mouse button has been pressed and has not yet been released.
         this._dragInProgress = false; // The drag has been started, and has not been dropped or cancelled yet.
         this._animationInProgress = false; // The drag is over and the item is in the process of animating to its original position (snapping back or reverting).
-
-        // During the drag, we eat enter/leave events so that actors don't prelight or show
-        // tooltips. But we remember the actors that we first left/last entered so we can
-        // fix up the hover state after the drag ends.
-        this._firstLeaveActor = null;
-        this._lastEnterActor = null;
+        this._dragCancellable = true;
 
         this._eventsGrabbed = false;
     },
 
     _onButtonPress : function (actor, event) {
-        if (!this.inhibit) {
-            if (event.get_button() != 1)
-                return false;
+        if (this.inhibit)
+            return false;
 
-            if (Tweener.getTweenCount(actor))
-                return false;
+        if (event.get_button() != 1)
+            return false;
 
-            this._buttonDown = true;
-            // special case St.Button: grabbing the pointer would mess up the
-            // internal state, so we start the drag manually on hover change
-            if (this.actor instanceof St.Button)
-                this.actor.connect('notify::hover',
-                                   Lang.bind(this, this._onButtonHoverChanged));
-            else
-                this._grabActor();
+        if (Tweener.getTweenCount(actor))
+            return false;
 
-            let [stageX, stageY] = event.get_coords();
-            this._dragStartX = stageX;
-            this._dragStartY = stageY;
-        }
+        this._buttonDown = true;
+        this._grabActor();
+
+        let [stageX, stageY] = event.get_coords();
+        this._dragStartX = stageX;
+        this._dragStartY = stageY;
 
         return false;
     },
 
-    _onButtonHoverChanged: function(button) {        
-        if (button.hover || !button.pressed)
-            return;
-
-        button.fake_release();
-        this.startDrag(this._dragStartX, this._dragStartY,
-                       global.get_current_time());        
-    },
-
     _grabActor: function() {
         Clutter.grab_pointer(this.actor);
         this._onEventId = this.actor.connect('event',
@@ -170,51 +154,36 @@ _Draggable.prototype = {
     },
 
     _ungrabActor: function() {
-        Clutter.ungrab_pointer();
         if (!this._onEventId)
             return;
+
+        Clutter.ungrab_pointer();
         this.actor.disconnect(this._onEventId);
         this._onEventId = null;
     },
 
     _grabEvents: function() {
         if (!this._eventsGrabbed) {
-            let eha = _getEventHandlerActor();
-            Clutter.grab_pointer(eha);
-            this.previousKeyFocusActor = global.stage.get_key_focus();
-            eha.grab_key_focus();
-            this._eventsGrabbed = true;
+            this._eventsGrabbed = Main.pushModal(_getEventHandlerActor());
+            if (this._eventsGrabbed)
+                Clutter.grab_pointer(_getEventHandlerActor());
         }
     },
 
     _ungrabEvents: function() {
         if (this._eventsGrabbed) {
-            if (this.previousKeyFocusActor) {
-                this.previousKeyFocusActor.grab_key_focus();
-            }
             Clutter.ungrab_pointer();
+            Main.popModal(_getEventHandlerActor());
             this._eventsGrabbed = false;
         }
     },
 
     _onEvent: function(actor, event) {
-        // prevent the watchdog timer from firing, for a while
-        this._setTimer(true);
-        
-        // Intercept BUTTON_PRESS to try to address a drag in progress condition 'dragging'
-        // on interminably - you started dragging, went off the panels, released the mouse
-        // button (which we can't track when you're not over a panel) then went back
-        // over a panel - you're in an in between state, no longer wanting to drag, but
-        // we still think you are.  This will cancel the drag if it senses a button press,
-        // assuming we'd never complete a drag with a press, only a release.
-        //
         // We intercept BUTTON_RELEASE event to know that the button was released in case we
         // didn't start the drag, to drop the draggable in case the drag was in progress, and
         // to complete the drag and ensure that whatever happens to be under the pointer does
         // not get triggered if the drag was cancelled with Esc.
-        if (event.type() == Clutter.EventType.BUTTON_PRESS && this._dragInProgress) {
-            this._cancelDrag(event.get_time());
-        } else if (event.type() == Clutter.EventType.BUTTON_RELEASE) {
+        if (event.type() == Clutter.EventType.BUTTON_RELEASE) {
             this._buttonDown = false;
             if (this._dragInProgress) {
                 return this._dragActorDropped(event);
@@ -235,38 +204,32 @@ _Draggable.prototype = {
             } else if (this._dragActor == null) {
                 return this._maybeStartDrag(event);
             }
-        } else if (event.type() == Clutter.EventType.LEAVE) {
-            if (this._firstLeaveActor == null)
-                this._firstLeaveActor = event.get_source();
-        } else if (event.type() == Clutter.EventType.ENTER) {
-            this._lastEnterActor = event.get_source();
-        } else if (event.type() == Clutter.EventType.KEY_PRESS) {
-            if (event.get_key_symbol() === Clutter.Escape) {
+        // We intercept KEY_PRESS event so that we can process Esc key press to cancel
+        // dragging and ignore all other key presses.
+        } else if (event.type() == Clutter.EventType.KEY_PRESS && this._dragInProgress) {
+            let symbol = event.get_key_symbol();
+            if (symbol === Clutter.KEY_Escape) {
                 this._cancelDrag(event.get_time());
+                return true;
             }
-            return true; // swallow all keyboard input during drag
         }
 
         return false;
     },
 
-    _setTimer: function(renew) {
-        if (this._dragEventTimeoutId) {
-            Mainloop.source_remove(this._dragEventTimeoutId);
-            this._dragEventTimeoutId = 0;
-        }
-        if (renew) {
-            this._dragEventTimeoutId = Mainloop.timeout_add(DRAG_EVENT_TIMEOUT_MS, 
-                Lang.bind(this, function() {
-                    if (this._dragInProgress) {
-                        this._cancelDrag();
-                        // _dragComplete is not always called by _cancelDrag.
-                        this._dragComplete();
-                    }
-                }));
-        }
+    /**
+     * fakeRelease:
+     *
+     * Fake a release event.
+     * Must be called if you want to intercept release events on draggable
+     * actors for other purposes (for example if you're using
+     * PopupMenu.ignoreRelease())
+     */
+    fakeRelease: function() {
+        this._buttonDown = false;
+        this._ungrabActor();
     },
-    
+
     /**
      * startDrag:
      * @stageX: X coordinate of event
@@ -278,11 +241,16 @@ _Draggable.prototype = {
      * for the draggable.
      */
     startDrag: function (stageX, stageY, time) {
-        this._setTimer(true); // activate the watchdog timer
-
         currentDraggable = this;
         this._dragInProgress = true;
 
+        // Special-case St.Button: the pointer grab messes with the internal
+        // state, so force a reset to a reasonable state here
+        if (this.actor instanceof St.Button) {
+            this.actor.fake_release();
+            this.actor.hover = false;
+        }
+
         this.emit('drag-begin', time);
         if (this._onEventId)
             this._ungrabActor();
@@ -293,7 +261,11 @@ _Draggable.prototype = {
         this._dragY = this._dragStartY = stageY;
 
         if (this.actor._delegate && this.actor._delegate.getDragActor) {
-            this._dragActor = this.actor._delegate.getDragActor(this._dragStartX, this._dragStartY);
+            this._dragActor = this.actor._delegate.getDragActor();
+            global.reparentActor(this._dragActor, Main.uiGroup);
+            this._dragActor.raise_top();
+            Cinnamon.util_set_hidden_from_pick(this._dragActor, true);
+
             // Drag actor does not always have to be the same as actor. For example drag actor
             // can be an image that's part of the actor. So to perform "snap back" correctly we need
             // to know what was the drag actor source.
@@ -322,27 +294,28 @@ _Draggable.prototype = {
             this._dragOffsetY = this._dragActor.y - this._dragStartY;
         } else {
             this._dragActor = this.actor;
+
             this._dragActorSource = undefined;
             this._dragOrigParent = this.actor.get_parent();
             this._dragOrigX = this._dragActor.x;
             this._dragOrigY = this._dragActor.y;
             this._dragOrigScale = this._dragActor.scale_x;
 
+            // Set the actor's scale such that it will keep the same
+            // transformed size when it's reparented to the uiGroup
+            let [scaledWidth, scaledHeight] = this.actor.get_transformed_size();
+            this._dragActor.set_scale(scaledWidth / this.actor.width,
+                                      scaledHeight / this.actor.height);
+
             let [actorStageX, actorStageY] = this.actor.get_transformed_position();
             this._dragOffsetX = actorStageX - this._dragStartX;
             this._dragOffsetY = actorStageY - this._dragStartY;
 
-            // Set the actor's scale such that it will keep the same
-            // transformed size when it's reparented to the uiGroup
-            let [scaledWidth, scaledHeight] = this.actor.get_transformed_size();
-            this.actor.set_scale(scaledWidth / this.actor.width,
-                                 scaledHeight / this.actor.height);
+            global.reparentActor(this._dragActor, Main.uiGroup);
+            this._dragActor.raise_top();
+            Cinnamon.util_set_hidden_from_pick(this._dragActor, true);
         }
 
-        this._dragActor.reparent(Main.uiGroup);
-        this._dragActor.raise_top();
-        Cinnamon.util_set_hidden_from_pick(this._dragActor, true);
-
         this._dragOrigOpacity = this._dragActor.opacity;
         if (this._dragActorOpacity != undefined)
             this._dragActor.opacity = this._dragActorOpacity;
@@ -375,8 +348,7 @@ _Draggable.prototype = {
                                        let currentScale = this._dragActor.scale_x / origScale;
                                        this._dragOffsetX = currentScale * origDragOffsetX;
                                        this._dragOffsetY = currentScale * origDragOffsetY;
-                                       this._dragActor.set_position(this._dragX + this._dragOffsetX,
-                                                                    this._dragY + this._dragOffsetY);
+                                       this._setDragActorPosition();
                                    },
                                    onUpdateScope: this });
             }
@@ -397,93 +369,111 @@ _Draggable.prototype = {
         return true;
     },
 
-    _setCursor:  function(result) {
-        if (result !== DragMotionResult.CONTINUE) {
-            try {
-                let cursor = DRAG_CURSOR_MAP[result];
-                global.set_cursor(cursor);
+    _updateDragHover : function () {
+        this._updateHoverId = 0;
+        let target = null;
+        let result = null;
+
+        let x = this._overrideX == undefined ? this._dragX : this._overrideX;
+        let y = this._overrideY == undefined ? this._dragY : this._overrideY;
+
+        if (this.recentDropTarget) {
+            let allocation = Cinnamon.util_get_transformed_allocation(this.recentDropTarget);
+
+            if (x < allocation.x1 || x > allocation.x2 || y < allocation.y1 || y > allocation.y2) {
+                this.recentDropTarget._delegate.handleDragOut();
+                this.recentDropTarget = null;
             }
-            catch (e) {
-                global.logError("bad DragMotionResult: " + result, e);
-            }
-            return true;
         }
+
+        if (this.target) {
+            target = this.target;
+        } else {
+            let stage = this._dragActor.get_stage();
+            if (!stage) {
+                return;
+            }
+            target = stage.get_actor_at_pos(Clutter.PickMode.ALL, x, y);
+        }
+
+        let dragEvent = {
+            x: this._dragX,
+            y: this._dragY,
+            dragActor: this._dragActor,
+            source: this.actor._delegate,
+            targetActor: target
+        };
+        for (let i = 0; i < dragMonitors.length; i++) {
+            let motionFunc = dragMonitors[i].dragMotion;
+            if (motionFunc) {
+                result = motionFunc(dragEvent);
+                if (result == DragMotionResult.MOVE_DROP || result == DragMotionResult.COPY_DROP) {
+                    break;
+                }
+            }
+        }
+
+        while (target) {
+            if (target._delegate && target._delegate.handleDragOver) {
+                let [r, targX, targY] = target.transform_stage_point(x, y);
+                // We currently loop through all parents on drag-over even if one of the children has handled it.
+                // We can check the return value of the function and break the loop if it's true if we don't want
+                // to continue checking the parents.
+                result = target._delegate.handleDragOver(this.actor._delegate,
+                                                             this._dragActor,
+                                                             targX,
+                                                             targY,
+                                                             0);
+                if (result == DragMotionResult.MOVE_DROP || result == DragMotionResult.COPY_DROP) {
+                    if (target._delegate.handleDragOut) this.recentDropTarget = target;
+                    break;
+                }
+            }
+            target = target.get_parent();
+        }
+        if (result in DRAG_CURSOR_MAP) global.set_cursor(DRAG_CURSOR_MAP[result]);
+        else global.set_cursor(Cinnamon.Cursor.DND_IN_DRAG);
         return false;
     },
-    
+
+    _queueUpdateDragHover: function() {
+        if (this._updateHoverId)
+            return;
+
+        this._updateHoverId = GLib.idle_add(GLib.PRIORITY_DEFAULT,
+                                            Lang.bind(this, this._updateDragHover));
+    },
+
     _updateDragPosition : function (event) {
         let [stageX, stageY] = event.get_coords();
         this._dragX = stageX;
         this._dragY = stageY;
 
-        // If we are dragging, update the position
-        if (this._dragActor) {
-            this._dragActor.set_position(stageX + this._dragOffsetX,
-                                         stageY + this._dragOffsetY);
-
-            let target = this._dragActor.get_stage().get_actor_at_pos(Clutter.PickMode.ALL,
-                                                                      stageX, stageY);
-
-            // We call observers only once per motion with the innermost
-            // target actor. If necessary, the observer can walk the
-            // parent itself.
-            let dragEvent = {
-                x: stageX,
-                y: stageY,
-                dragActor: this._dragActor,
-                source: this.actor._delegate,
-                targetActor: target
-            };
-            for (let i = 0; i < dragMonitors.length; i++) {
-                let motionFunc = dragMonitors[i].dragMotion;
-                if (motionFunc) {
-                    let result = motionFunc(dragEvent);
-                    if (this._setCursor(result)) {
-                        return true;
-                    }
-                }
-            }
-
-            if (this.target) {
-                if (this.target._delegate && this.target._delegate.handleDragOver){
-                    let [r, targX, targY] = this.target.transform_stage_point(stageX, stageY);
-                    let result = this.target._delegate.handleDragOver(this.actor._delegate,
-                                                                      this._dragActor,
-                                                                      targX,
-                                                                      targY,
-                                                                      event.get_time());
-                    if (this._setCursor(result)) {
-                        return true;
-                    }
-                }
-            }
-
-            while (target) {
-                if (target._delegate && target._delegate.handleDragOver) {
-                    let [r, targX, targY] = target.transform_stage_point(stageX, stageY);
-                    // We currently loop through all parents on drag-over even if one of the children has handled it.
-                    // We can check the return value of the function and break the loop if it's true if we don't want
-                    // to continue checking the parents.
-                    let result = target._delegate.handleDragOver(this.actor._delegate,
-                                                                 this._dragActor,
-                                                                 targX,
-                                                                 targY,
-                                                                 event.get_time());
-                    if (this._setCursor(result)) {
-                        return true;
-                    }
-                }
-                target = target.get_parent();
-            }
-            global.set_cursor(Cinnamon.Cursor.DND_IN_DRAG);
-        }
+        this._setDragActorPosition();
 
+        this._queueUpdateDragHover();
         return true;
     },
 
+    _setDragActorPosition: function() {
+        this._dragActor.x = this._overrideX == undefined ?
+            this._dragX + this._dragOffsetX : this._overrideX;
+
+        this._dragActor.y = this._overrideY == undefined ?
+            this._dragY + this._dragOffsetY : this._overrideY;
+    },
+
     _dragActorDropped: function(event) {
         let [dropX, dropY] = event.get_coords();
-        let target = this._dragActor.get_stage().get_actor_at_pos(Clutter.PickMode.ALL,
+        let target = null;
+
+        if (this._overrideX != undefined) dropX = this._overrideX;
+        if (this._overrideY != undefined) dropY = this._overrideY;
+
+        if (this.target)
+            target = this.target;
+        else
+            target = this._dragActor.get_stage().get_actor_at_pos(Clutter.PickMode.ALL,
                                                                   dropX, dropY);
 
         // We call observers only once per motion with the innermost
@@ -506,34 +496,10 @@ _Draggable.prototype = {
                 }
         }
 
-        if (this.target) {
-            if (this.target._delegate && this.target._delegate.acceptDrop){
-                let [r, targX, targY] = this.target.transform_stage_point(dropX, dropY);
-                if (this.target._delegate.acceptDrop(this.actor._delegate,
-                                                     this._dragActor,
-                                                     targX,
-                                                     targY,
-                                                     event.get_time())) {
-                    if (this._actorDestroyed)
-                        return true;
-                    // If it accepted the drop without taking the actor,
-                    // handle it ourselves.
-                    if (this._dragActor.get_parent() == Main.uiGroup) {
-                        if (this._restoreOnSuccess) {
-                            this._restoreDragActor(event.get_time());
-                            return true;
-                        } else
-                            this._dragActor.destroy();
-                    }
-
-                    this._dragInProgress = false;
-                    global.unset_cursor();
-                    this.emit('drag-end', event.get_time(), true);
-                    this._dragComplete();
-                    return true;
-                }
-            }
-        }
+        // At this point it is too late to cancel a drag by destroying
+        // the actor, the fate of which is decided by acceptDrop and its
+        // side-effects
+        this._dragCancellable = false;
 
         while (target) {
             if (target._delegate && target._delegate.acceptDrop) {
@@ -543,11 +509,9 @@ _Draggable.prototype = {
                                                 targX,
                                                 targY,
                                                 event.get_time())) {
-                    if (this._actorDestroyed)
-                        return true;
                     // If it accepted the drop without taking the actor,
                     // handle it ourselves.
-                    if (this._dragActor.get_parent() == Main.uiGroup) {
+                    if (!this._dragActor.is_finalized() && this._dragActor.get_parent() === Main.uiGroup) {
                         if (this._restoreOnSuccess) {
                             this._restoreDragActor(event.get_time());
                             return true;
@@ -602,9 +566,7 @@ _Draggable.prototype = {
         return [x, y, scale];
     },
 
-    _cancelDrag: function(eventTimeOpt) {
-        this._setTimer(false);
-        let eventTime = eventTimeOpt || global.get_current_time();
+    _cancelDrag: function(eventTime) {
         this.emit('drag-cancelled', eventTime);
         this._dragInProgress = false;
         let [snapBackX, snapBackY, snapBackScale] = this._getRestoreLocation();
@@ -614,7 +576,7 @@ _Draggable.prototype = {
             if (!this._buttonDown)
                 this._dragComplete();
             this.emit('drag-end', eventTime, false);
-            if (!this._dragOrigParent && this._dragActor)
+            if (!this._dragOrigParent)
                 this._dragActor.destroy();
 
             return;
@@ -638,7 +600,7 @@ _Draggable.prototype = {
 
     _restoreDragActor: function(eventTime) {
         this._dragInProgress = false;
-        [restoreX, restoreY, restoreScale] = this._getRestoreLocation();
+        let [restoreX, restoreY, restoreScale] = this._getRestoreLocation();
 
         // fade the actor back in at its original location
         this._dragActor.set_position(restoreX, restoreY);
@@ -658,7 +620,7 @@ _Draggable.prototype = {
 
     _onAnimationComplete : function (dragActor, eventTime) {
         if (this._dragOrigParent) {
-            dragActor.reparent(this._dragOrigParent);
+            global.reparentActor (dragActor, this._dragOrigParent);
             dragActor.set_scale(this._dragOrigScale, this._dragOrigScale);
             dragActor.set_position(this._dragOrigX, this._dragOrigY);
         } else {
@@ -672,39 +634,22 @@ _Draggable.prototype = {
             this._dragComplete();
     },
 
-    // Actor is an actor we have entered or left during the drag; call
-    // st_widget_sync_hover on all StWidget ancestors
-    _syncHover: function(actor) {
-        while (actor) {
-            let parent = actor.get_parent();
-            if (actor instanceof St.Widget)
-                actor.sync_hover();
-
-            actor = parent;
-        }
-    },
-
     _dragComplete: function() {
-        this._setTimer(false);
-        if (!this._actorDestroyed)
+        if (!this._actorDestroyed && !this._dragActor.is_finalized())
             Cinnamon.util_set_hidden_from_pick(this._dragActor, false);
 
         this._ungrabEvents();
+        global.sync_pointer();
 
-        if (this._firstLeaveActor) {
-            this._syncHover(this._firstLeaveActor);
-            this._firstLeaveActor = null;
-        }
-
-        if (this._lastEnterActor) {
-            this._syncHover(this._lastEnterActor);
-            this._lastEnterActor = null;
+        if (this._updateHoverId) {
+            GLib.source_remove(this._updateHoverId);
+            this._updateHoverId = 0;
         }
 
         this._dragActor = undefined;
         currentDraggable = null;
     }
-};
+});
 
 Signals.addSignalMethods(_Draggable.prototype);
 
@@ -712,11 +657,8 @@ Signals.addSignalMethods(_Draggable.prototype);
  * makeDraggable:
  * @actor: Source actor
  * @params: (optional) Additional parameters
- * @target: (optional) Actor that has priority as an accepting target
  *
  * Create an object which controls drag and drop for the given actor.
- * @target has priority when finding a target to accept the drag
- * actor. This can be used when the target cannot be reached.
  *
  * If %manualMode is %true in @params, do not automatically start
  * drag and drop on click
@@ -807,29 +749,42 @@ GenericDragItemContainer.prototype = {
         if (this.child == actor)
             return;
 
-        this.actor.destroy_children();
+        this.actor.destroy_all_children();
 
         this.child = actor;
         this.actor.add_actor(this.child);
     },
 
-    animateIn: function() {
-        if (this.child == null)
+    animateIn: function(onCompleteFunc) {
+        if (!this.child) {
+            if (typeof(onCompleteFunc) === 'function')
+                onCompleteFunc();
             return;
+        }
 
         this.childScale = 0;
         this.childOpacity = 0;
-        Tweener.addTween(this,
-                         { childScale: 1.0,
-                           childOpacity: 255,
-                           time: DND_ANIMATION_TIME,
-                           transition: 'easeOutQuad'
-                         });
+
+        let params = { childScale: 1.0,
+                       childOpacity: 255,
+                       time: DND_ANIMATION_TIME,
+                       transition: 'easeOutQuad' };
+
+        if (typeof(onCompleteFunc) === 'function')
+            params.onComplete = onCompleteFunc;
+
+        Tweener.addTween(this, params);
     },
 
-    animateOutAndDestroy: function() {
-        if (this.child == null) {
+    animateOutAndDestroy: function(onCompleteFunc) {
+        let _onComplete = () => {
+            if (typeof(onCompleteFunc) === 'function')
+                onCompleteFunc();
             this.actor.destroy();
+        };
+
+        if (!this.child) {
+            _onComplete();
             return;
         }
 
@@ -840,13 +795,11 @@ GenericDragItemContainer.prototype = {
                            childOpacity: 0,
                            time: DND_ANIMATION_TIME,
                            transition: 'easeOutQuad',
-                           onComplete: Lang.bind(this, function() {
-                               this.actor.destroy();
-                           })
-                         });
+                           onComplete: _onComplete });
     },
 
     set childScale(scale) {
+        if (this.child.is_finalized()) return;
         this._childScale = scale;
 
         if (this.child == null)
@@ -862,6 +815,7 @@ GenericDragItemContainer.prototype = {
     },
 
     set childOpacity(opacity) {
+        if (this.child.is_finalized()) return;
         this._childOpacity = opacity;
 
         if (this.child == null)
@@ -888,3 +842,14 @@ GenericDragPlaceholderItem.prototype = {
         this.setChild(new St.Bin({ style_class: 'drag-placeholder' }));
     }
 };
+
+var LauncherDraggable = class {
+    constructor(launchersBox) {
+        this.launchersBox = launchersBox;
+    }
+
+    getId() {
+        /* Implemented by draggable launchers */
+        global.logError("Could not complete drag-and-drop.  Launcher does not implement LauncherDraggable");
+    }
+};
