@@ -19,8 +19,7 @@
 
 #include "cinnamon-window-tracker-private.h"
 #include "cinnamon-app-private.h"
-#include "cinnamon-global.h"
-#include "cinnamon-marshal.h"
+#include "cinnamon-global-private.h"
 #include "st.h"
 
 /* This file includes modified code from
@@ -44,6 +43,8 @@ struct _CinnamonWindowTracker
 {
   GObject parent;
 
+  CinnamonGlobal *global;
+
   CinnamonApp *focus_app;
 
   /* <MetaWindow * window, CinnamonApp *app> */
@@ -62,8 +63,7 @@ enum {
 
 enum {
   STARTUP_SEQUENCE_CHANGED,
-  TRACKED_WINDOWS_CHANGED,
-
+  WINDOW_APP_CHANGED,
   LAST_SIGNAL
 };
 
@@ -117,44 +117,21 @@ cinnamon_window_tracker_class_init (CinnamonWindowTrackerClass *klass)
                                    CINNAMON_TYPE_WINDOW_TRACKER,
                                    G_SIGNAL_RUN_LAST,
                                    0,
-                                   NULL, NULL,
-                                   g_cclosure_marshal_VOID__BOXED,
+                                   NULL, NULL, NULL,
                                    G_TYPE_NONE, 1, CINNAMON_TYPE_STARTUP_SEQUENCE);
-  signals[TRACKED_WINDOWS_CHANGED] = g_signal_new ("tracked-windows-changed",
-                                                   CINNAMON_TYPE_WINDOW_TRACKER,
-                                                   G_SIGNAL_RUN_LAST,
-                                                   0,
-                                                   NULL, NULL,
-                                                   g_cclosure_marshal_VOID__VOID,
-                                                   G_TYPE_NONE, 0);
-}
 
-/**
- * get_appid_from_window:
- *
- * Turn the WM_CLASS property into our best guess at a .desktop file id.
- */
-static char *
-get_appid_from_window (MetaWindow  *window)
-{
-  const char *wmclass;
-  char *appid_guess;
-
-  wmclass = meta_window_get_wm_class (window);
-  if (!wmclass)
-    return NULL;
-
-  appid_guess = g_ascii_strdown (wmclass, -1);
-
-  /* This handles "Fedora Eclipse", probably others.
-   * Note g_strdelimit is modify-in-place. */
-  g_strdelimit (appid_guess, " ", '-');
-
-  return appid_guess;
+  signals[WINDOW_APP_CHANGED] = g_signal_new ("window-app-changed",
+                                              CINNAMON_TYPE_WINDOW_TRACKER,
+                                              G_SIGNAL_RUN_LAST,
+                                              0,
+                                              NULL, NULL, NULL,
+                                              G_TYPE_NONE, 1, META_TYPE_WINDOW);
 }
 
 /**
  * cinnamon_window_tracker_is_window_interesting:
+ * @tracker: the CinnamonWindowTracker
+ * @window: a #MetaWindow
  *
  * The CinnamonWindowTracker associates certain kinds of windows with
  * applications; however, others we don't want to
@@ -166,42 +143,55 @@ get_appid_from_window (MetaWindow  *window)
  * exclude other window types like tooltip explicitly, though generally
  * most of these should be override-redirect.
  *
- * Returns: %TRUE iff a window is "interesting"
+ * Returns: %TRUE if a window is "interesting"
  */
 gboolean
-cinnamon_window_tracker_is_window_interesting (MetaWindow *window)
+cinnamon_window_tracker_is_window_interesting (CinnamonWindowTracker *tracker, MetaWindow *window)
 {
-  if (meta_window_is_override_redirect (window)
-      || meta_window_is_skip_taskbar (window))
-    return FALSE;
+  return meta_window_is_interesting (window);
+}
 
-  switch (meta_window_get_window_type (window))
+static gboolean
+is_window_only_browser_app (MetaWindow  *window,
+                            const gchar *wm_instance,
+                            const gchar *wm_class)
+{
+    if (wm_instance == NULL || wm_class == NULL)
     {
-      /* Definitely ignore these. */
-      case META_WINDOW_DESKTOP:
-      case META_WINDOW_DOCK:
-      case META_WINDOW_SPLASHSCREEN:
-      /* Should have already been handled by override_redirect above,
-       * but explicitly list here so we get the "unhandled enum"
-       * warning if in the future anything is added.*/
-      case META_WINDOW_DROPDOWN_MENU:
-      case META_WINDOW_POPUP_MENU:
-      case META_WINDOW_TOOLTIP:
-      case META_WINDOW_NOTIFICATION:
-      case META_WINDOW_COMBO:
-      case META_WINDOW_DND:
-      case META_WINDOW_OVERRIDE_OTHER:
         return FALSE;
-      case META_WINDOW_NORMAL:
-      case META_WINDOW_DIALOG:
-      case META_WINDOW_MODAL_DIALOG:
-      case META_WINDOW_MENU:
-      case META_WINDOW_TOOLBAR:
-      case META_WINDOW_UTILITY:
-        break;
     }
 
-  return TRUE;
+    /* This is to catch browser-based app windows that are spawned
+     * windows (such as from clicking an extension - LINE is a test
+     * case). There's no desktop file but we don't want to use the
+     * browser icon and the window has a pixmap to use. */
+
+    g_autofree gchar *utf8_instance = NULL;
+    g_autofree gchar *utf8_class = NULL;
+    g_autofree gchar *lower_instance = NULL;
+    g_autofree gchar *lower_class = NULL;
+
+    utf8_instance = g_utf8_make_valid (wm_instance, -1);
+    utf8_class = g_utf8_make_valid (wm_class, -1);
+
+    lower_instance = g_utf8_casefold (utf8_instance, -1);
+    lower_class = g_utf8_casefold (utf8_class, -1);
+
+    // If the instance and instance are the same ["chromium-browser", "Chromium-browser"],
+    // don't consider it special.
+    if (g_strcmp0 (lower_instance, lower_class) == 0)
+    {
+        return FALSE;
+    }
+
+    // If the two strings differ, and one has browser in it, then assume it's a web app.
+    if (g_strstr_len (lower_instance, -1, "browser") || g_strstr_len (lower_class, -1, "browser"))
+    {
+        return TRUE;
+    }
+
+    // we don't know, match some other way.
+    return FALSE;
 }
 
 /**
@@ -218,23 +208,115 @@ get_app_from_window_wmclass (MetaWindow  *window)
 {
   CinnamonApp *app;
   CinnamonAppSystem *appsys;
-  char *wmclass;
-  char *with_desktop;
+  const char *wm_class;
+  const char *wm_instance;
 
   appsys = cinnamon_app_system_get_default ();
-  wmclass = get_appid_from_window (window);
 
-  if (!wmclass)
+  /* Notes on the heuristics used here:
+     much of the complexity here comes from the desire to support
+     Chrome apps.
+
+     From https://bugzilla.gnome.org/show_bug.cgi?id=673657#c13
+
+     Currently chrome sets WM_CLASS as follows (the first string is the 'instance',
+     the second one is the 'class':
+
+     For the normal browser:
+     WM_CLASS(STRING) = "chromium", "Chromium"
+
+     For a bookmarked page (through 'Tools -> Create application shortcuts')
+     WM_CLASS(STRING) = "wiki.gnome.org__GnomeShell_ApplicationBased", "Chromium"
+
+     For an application from the chrome store (with a .desktop file created through
+     right click, "Create shortcuts" from Chrome's apps overview)
+     WM_CLASS(STRING) = "crx_blpcfgokakmgnkcojhhkbfbldkacnbeo", "Chromium"
+
+     The .desktop file has a matching StartupWMClass, but the name differs, e.g. for
+     the store app (youtube) there is
+
+     .local/share/applications/chrome-blpcfgokakmgnkcojhhkbfbldkacnbeo-Default.desktop
+
+     with
+
+     StartupWMClass=crx_blpcfgokakmgnkcojhhkbfbldkacnbeo
+
+     Note that chromium (but not google-chrome!) includes a StartupWMClass=chromium
+     in their .desktop file, so we must match the instance first.
+
+     Also note that in the good case (regular gtk+ app without hacks), instance and
+     class are the same except for case and there is no StartupWMClass at all.
+  */
+
+#if 0
+  wm_instance = meta_window_get_wm_class_instance (window);
+  wm_class = meta_window_get_wm_class (window);
+  g_printerr ("get_app_from_window_wmclass: wm_instance: %s, wm_class: %s\n", wm_instance, wm_class);
+#endif
+
+  /* first try a match from WM_CLASS (instance part) to StartupWMClass */
+  wm_instance = meta_window_get_wm_class_instance (window);
+  app = cinnamon_app_system_lookup_startup_wmclass (appsys, wm_instance);
+  if (app != NULL)
+    return g_object_ref (app);
+
+  /* then try a match from WM_CLASS to StartupWMClass */
+  wm_class = meta_window_get_wm_class (window);
+  app = cinnamon_app_system_lookup_startup_wmclass (appsys, wm_class);
+  if (app != NULL)
+    return g_object_ref (app);
+
+  /* then try a match from WM_CLASS (instance part) to .desktop */
+  app = cinnamon_app_system_lookup_desktop_wmclass (appsys, wm_instance);
+  if (app != NULL)
+    return g_object_ref (app);
+
+  /* figure out if it's a browser app without a desktop file. */
+  if (is_window_only_browser_app (window, wm_instance, wm_class))
+  {
+    // Force it to use the window pixmap.
+    return NULL;
+  }
+
+  /* finally, try a match from WM_CLASS to .desktop */
+  app = cinnamon_app_system_lookup_desktop_wmclass (appsys, wm_class);
+  if (app != NULL)
+    return g_object_ref (app);
+
+  return NULL;
+}
+
+/**
+ * get_app_from_gapplication_id:
+ * @monitor: a #CinnamonWindowTracker
+ * @window: a #MetaWindow
+ *
+ * Looks only at the given window, and attempts to determine
+ * an application based on _GTK_APPLICATION_ID.  If one can't be determined,
+ * return %NULL.
+ *
+ * Return value: (transfer full): A newly-referenced #CinnamonApp, or %NULL
+ */
+static CinnamonApp *
+get_app_from_gapplication_id (MetaWindow  *window)
+{
+  CinnamonApp *app;
+  CinnamonAppSystem *appsys;
+  const char *id;
+  char *desktop_file;
+
+  appsys = cinnamon_app_system_get_default ();
+
+  id = meta_window_get_gtk_application_id (window);
+  if (!id)
     return NULL;
 
-  with_desktop = g_strjoin (NULL, wmclass, ".desktop", NULL);
-  g_free (wmclass);
-
-  app = cinnamon_app_system_lookup_heuristic_basename (appsys, with_desktop);
-  if (app != NULL)
+  desktop_file = g_strconcat (id, ".desktop", NULL);
+  app = cinnamon_app_system_lookup_app (appsys, desktop_file);
+  if (app)
     g_object_ref (app);
-  g_free (with_desktop);
 
+  g_free (desktop_file);
   return app;
 }
 
@@ -318,6 +400,85 @@ get_app_from_window_pid (CinnamonWindowTracker  *tracker,
   return result;
 }
 
+static CinnamonApp *
+get_app_for_flatpak_window (MetaWindow *window)
+{
+  CinnamonAppSystem *appsys;
+  CinnamonApp *app = NULL;
+  CinnamonApp *result = NULL;
+  gchar *info_filename;
+  GFile *file;
+  int pid = meta_window_get_client_pid (window);
+
+  g_return_val_if_fail (pid > 0, NULL);
+
+  info_filename = g_strdup_printf ("/proc/%u/root/.flatpak-info", pid);
+  file = g_file_new_for_path (info_filename);
+
+  if (g_file_query_exists (file, NULL)) {
+    gchar *wm_class;
+    gchar *wm_instance;
+    GKeyFile *keyfile;
+
+    appsys = cinnamon_app_system_get_default ();
+
+    keyfile = g_key_file_new ();
+    if (g_key_file_load_from_file (keyfile, info_filename, G_KEY_FILE_NONE, NULL))
+    {
+      gchar *app_id;
+      app_id = g_key_file_get_string (keyfile, "Application", "name", NULL);
+      app = cinnamon_app_system_lookup_flatpak_app_id (appsys, app_id);
+
+      if (app != NULL) {
+        result = g_object_ref (app);
+      }
+    }
+    g_key_file_unref (keyfile);
+
+    wm_instance = g_strconcat(meta_window_get_wm_class_instance (window), GMENU_DESKTOPAPPINFO_FLATPAK_SUFFIX, NULL);
+    wm_class = g_strconcat(meta_window_get_wm_class (window), GMENU_DESKTOPAPPINFO_FLATPAK_SUFFIX, NULL);
+
+    if (result == NULL) {
+      app = cinnamon_app_system_lookup_startup_wmclass (appsys, wm_instance);
+      if (app != NULL) {
+        result = g_object_ref (app);
+      }
+    }
+
+    /* then try a match from WM_CLASS to StartupWMClass */
+    if (result == NULL) {
+      app = cinnamon_app_system_lookup_startup_wmclass (appsys, wm_class);
+      if (app != NULL) {
+        result = g_object_ref (app);
+      }
+    }
+
+    /* then try a match from WM_CLASS (instance part) to .desktop */
+    if (result == NULL) {
+      app = cinnamon_app_system_lookup_desktop_wmclass (appsys, wm_instance);
+      if (app != NULL) {
+        result = g_object_ref (app);
+      }
+    }
+
+    /* finally, try a match from WM_CLASS to .desktop */
+    if (result == NULL) {
+      app = cinnamon_app_system_lookup_desktop_wmclass (appsys, wm_class);
+      if (app != NULL) {
+        result = g_object_ref (app);
+      }
+    }
+
+    g_free (wm_instance);
+    g_free (wm_class);
+  }
+
+  g_free (info_filename);
+  g_object_unref(file);
+
+  return result;
+}
+
 /**
  * get_app_for_window:
  *
@@ -351,6 +512,18 @@ get_app_for_window (CinnamonWindowTracker    *tracker,
   if (meta_window_is_remote (window))
     return _cinnamon_app_new_for_window (window);
 
+  /* Check if the window was launched from a sandboxed app, e.g. Flatpak */
+  result = get_app_for_flatpak_window (window);
+  if (result != NULL)
+    return result;
+
+  /* Check if the window has a GApplication ID attached; this is
+   * canonical if it does
+   */
+  result = get_app_from_gapplication_id (window);
+  if (result != NULL)
+    return result;
+
   /* Check if the app's WM_CLASS specifies an app; this is
    * canonical if it does.
    */
@@ -398,22 +571,29 @@ get_app_for_window (CinnamonWindowTracker    *tracker,
   return result;
 }
 
-const char *
-_cinnamon_window_tracker_get_app_context (CinnamonWindowTracker *tracker, CinnamonApp *app)
-{
-  return "";
-}
-
 static void
 update_focus_app (CinnamonWindowTracker *self)
 {
   MetaWindow *new_focus_win;
   CinnamonApp *new_focus_app;
 
-  new_focus_win = meta_display_get_focus_window (cinnamon_global_get_display (cinnamon_global_get ()));
+  new_focus_win = meta_display_get_focus_window (self->global->meta_display);
   new_focus_app = new_focus_win ? cinnamon_window_tracker_get_window_app (self, new_focus_win) : NULL;
 
   set_focus_app (self, new_focus_app);
+  g_clear_object (&new_focus_app);
+}
+
+static void
+tracked_window_changed (CinnamonWindowTracker *self,
+                        MetaWindow         *window)
+{
+  /* It's simplest to just treat this as a remove + add. */
+  disassociate_window (self, window);
+  track_window (self, window);
+  /* also just recalculate the focused app, in case it was the focused
+     window that changed */
+  update_focus_app (self);
 }
 
 static void
@@ -422,13 +602,16 @@ on_wm_class_changed (MetaWindow  *window,
                      gpointer     user_data)
 {
   CinnamonWindowTracker *self = CINNAMON_WINDOW_TRACKER (user_data);
+  tracked_window_changed (self, window);
+}
 
-  /* It's simplest to just treat this as a remove + add. */
-  disassociate_window (self, window);
-  track_window (self, window);
-  /* also just recaulcuate the focused app, in case it was the focused
-     window that changed */
-  update_focus_app (self);
+static void
+on_gtk_application_id_changed (MetaWindow  *window,
+                               GParamSpec  *pspec,
+                               gpointer     user_data)
+{
+  CinnamonWindowTracker *self = CINNAMON_WINDOW_TRACKER (user_data);
+  tracked_window_changed (self, window);
 }
 
 static void
@@ -437,7 +620,7 @@ track_window (CinnamonWindowTracker *self,
 {
   CinnamonApp *app;
 
-  if (!cinnamon_window_tracker_is_window_interesting (window))
+  if (!meta_window_is_interesting (window))
     return;
 
   app = get_app_for_window (self, window);
@@ -448,10 +631,11 @@ track_window (CinnamonWindowTracker *self,
   g_hash_table_insert (self->window_to_app, window, app);
 
   g_signal_connect (window, "notify::wm-class", G_CALLBACK (on_wm_class_changed), self);
+  g_signal_connect (window, "notify::gtk-application-id", G_CALLBACK (on_gtk_application_id_changed), self);
 
   _cinnamon_app_add_window (app, window);
 
-  g_signal_emit (self, signals[TRACKED_WINDOWS_CHANGED], 0);
+  g_signal_emit (self, signals[WINDOW_APP_CHANGED], 0, window);
 }
 
 static void
@@ -478,13 +662,14 @@ disassociate_window (CinnamonWindowTracker   *self,
 
   g_hash_table_remove (self->window_to_app, window);
 
-  if (cinnamon_window_tracker_is_window_interesting (window))
+  if (meta_window_is_interesting (window))
     {
       _cinnamon_app_remove_window (app, window);
       g_signal_handlers_disconnect_by_func (window, G_CALLBACK(on_wm_class_changed), self);
+      g_signal_handlers_disconnect_by_func (window, G_CALLBACK (on_gtk_application_id_changed), self);
     }
 
-  g_signal_emit (self, signals[TRACKED_WINDOWS_CHANGED], 0);
+  g_signal_emit (self, signals[WINDOW_APP_CHANGED], 0, window);
 
   g_object_unref (app);
 }
@@ -501,7 +686,7 @@ static void
 load_initial_windows (CinnamonWindowTracker *tracker)
 {
   GList *workspaces, *iter;
-  MetaScreen *screen = cinnamon_global_get_screen (cinnamon_global_get ());
+  MetaScreen *screen = tracker->global->meta_screen;
   workspaces = meta_screen_get_workspaces (screen);
 
   for (iter = workspaces; iter; iter = iter->next)
@@ -556,7 +741,7 @@ static void
 init_window_tracking (CinnamonWindowTracker *self)
 {
   MetaDisplay *display;
-  MetaScreen *screen = cinnamon_global_get_screen (cinnamon_global_get ());
+  MetaScreen *screen = self->global->meta_screen;
 
   g_signal_connect (screen, "notify::n-workspaces",
                     G_CALLBACK (cinnamon_window_tracker_on_n_workspaces_changed), self);
@@ -586,12 +771,14 @@ cinnamon_window_tracker_init (CinnamonWindowTracker *self)
 {
   MetaScreen *screen;
 
+  self->global = cinnamon_global_get ();
+
   self->window_to_app = g_hash_table_new_full (g_direct_hash, g_direct_equal,
                                                NULL, (GDestroyNotify) g_object_unref);
 
   self->launched_pid_to_app = g_hash_table_new_full (NULL, NULL, NULL, (GDestroyNotify) g_object_unref);
 
-  screen = cinnamon_global_get_screen (cinnamon_global_get ());
+  screen = self->global->meta_screen;
 
   g_signal_connect (G_OBJECT (screen), "startup-sequence-changed",
                     G_CALLBACK (on_startup_sequence_changed), self);
@@ -647,7 +834,7 @@ cinnamon_window_tracker_get_window_app (CinnamonWindowTracker *tracker,
  * Returns: (transfer none): A #CinnamonApp, or %NULL if none
  */
 CinnamonApp *
-cinnamon_window_tracker_get_app_from_pid (CinnamonWindowTracker *self, 
+cinnamon_window_tracker_get_app_from_pid (CinnamonWindowTracker *self,
                                        int                 pid)
 {
   GSList *running = cinnamon_app_system_get_running (cinnamon_app_system_get_default());
@@ -749,9 +936,7 @@ on_focus_window_changed (MetaDisplay        *display,
 GSList *
 cinnamon_window_tracker_get_startup_sequences (CinnamonWindowTracker *self)
 {
-  CinnamonGlobal *global = cinnamon_global_get ();
-  MetaScreen *screen = cinnamon_global_get_screen (global);
-  return meta_screen_get_startup_sequences (screen);
+  return meta_screen_get_startup_sequences (self->global->meta_screen);
 }
 
 /* sn_startup_sequence_ref returns void, so make a
@@ -792,6 +977,7 @@ CinnamonApp *
 cinnamon_startup_sequence_get_app (CinnamonStartupSequence *sequence)
 {
   const char *appid;
+  char *basename;
   CinnamonAppSystem *appsys;
   CinnamonApp *app;
 
@@ -799,8 +985,10 @@ cinnamon_startup_sequence_get_app (CinnamonStartupSequence *sequence)
   if (!appid)
     return NULL;
 
+  basename = g_path_get_basename (appid);
   appsys = cinnamon_app_system_get_default ();
-  app = cinnamon_app_system_lookup_app_for_path (appsys, appid);
+  app = cinnamon_app_system_lookup_app (appsys, basename);
+  g_free (basename);
   return app;
 }
 
@@ -833,14 +1021,22 @@ cinnamon_startup_sequence_create_icon (CinnamonStartupSequence *sequence, guint
   icon_name = sn_startup_sequence_get_icon_name ((SnStartupSequence*)sequence);
   if (!icon_name)
     {
+      gint scale;
+      CinnamonGlobal *global;
+      StThemeContext *context;
+
       texture = clutter_texture_new ();
-      clutter_actor_set_size (texture, size, size);
+
+      global = cinnamon_global_get ();
+      context = st_theme_context_get_for_stage (global->stage);
+      g_object_get (context, "scale-factor", &scale, NULL);
+
+      clutter_actor_set_size (texture, size * scale, size * scale);
       return texture;
     }
 
   themed = g_themed_icon_new (icon_name);
-  texture = st_texture_cache_load_gicon (st_texture_cache_get_default (),
-                                         NULL, themed, size);
+  texture = g_object_new (ST_TYPE_ICON, "gicon", themed, "icon-size", size, NULL);
   g_object_unref (G_OBJECT (themed));
   return texture;
 }
@@ -852,7 +1048,7 @@ cinnamon_startup_sequence_create_icon (CinnamonStartupSequence *sequence, guint
  * Return Value: (transfer none): The global #CinnamonWindowTracker instance
  */
 CinnamonWindowTracker *
-cinnamon_window_tracker_get_default ()
+cinnamon_window_tracker_get_default (void)
 {
   static CinnamonWindowTracker *instance;
 
