@@ -59,7 +59,6 @@
  */
 
 #include "st-scroll-view.h"
-#include "st-marshal.h"
 #include "st-scroll-bar.h"
 #include "st-scrollable.h"
 #include "st-scroll-view-fade.h"
@@ -70,21 +69,14 @@ static void clutter_container_iface_init (ClutterContainerIface *iface);
 
 static ClutterContainerIface *st_scroll_view_parent_iface = NULL;
 
-G_DEFINE_TYPE_WITH_CODE (StScrollView, st_scroll_view, ST_TYPE_BIN,
-                         G_IMPLEMENT_INTERFACE (CLUTTER_TYPE_CONTAINER,
-                                                clutter_container_iface_init))
-
-#define SCROLL_VIEW_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), \
-                                                             ST_TYPE_SCROLL_VIEW, \
-                                                             StScrollViewPrivate))
-
 #define AUTO_SCROLL_POLL_INTERVAL 15
 
 #define AUTO_SCROLL_TOTAL_REGION 100
 
-#define AUTO_SCROLL_OVERLAP 10 /* autoscroll region extends this far into the scroll view
-                                * remaining portion of region is above or below the view
-                                */
+/* autoscroll region extends this far into the scroll view
+ * remaining portion of region is above or below the view
+ */
+#define AUTO_SCROLL_OVERLAP 10
 
 /* gdouble - from 0 to TOTAL_REGION / this number to get scroll delta */
 #define AUTO_SCROLL_SPEED_DIVISOR 4.0
@@ -109,17 +101,25 @@ struct _StScrollViewPrivate
   gfloat        row_size;
   gfloat        column_size;
 
+  GSettings *settings;
+  gint setting_connect_id;
+
   StScrollViewFade *vfade_effect;
 
-  gboolean      row_size_set : 1;
-  gboolean      column_size_set : 1;
+  guint         row_size_set : 1;
+  guint         column_size_set : 1;
   guint         mouse_scroll : 1;
   guint         hscrollbar_visible : 1;
   guint         vscrollbar_visible : 1;
-  gboolean      auto_scroll : 1;
-  guint         auto_scroll_timeout_id : 1;
+  guint         auto_scroll : 1;
+  guint         auto_scroll_timeout_id;
 };
 
+G_DEFINE_TYPE_WITH_CODE (StScrollView, st_scroll_view, ST_TYPE_BIN,
+                         G_ADD_PRIVATE (StScrollView)
+                         G_IMPLEMENT_INTERFACE (CLUTTER_TYPE_CONTAINER,
+                                                clutter_container_iface_init))
+
 enum {
   PROP_0,
 
@@ -166,6 +166,7 @@ st_scroll_view_get_property (GObject    *object,
       break;
     case PROP_AUTO_SCROLL:
       g_value_set_boolean (value, priv->auto_scroll);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
@@ -254,6 +255,8 @@ calculate_and_scroll (ClutterActor  *self,
                     "value", &vvalue,
                     NULL);
     st_adjustment_set_value (priv->vadjustment, vvalue + delta);
+
+    clutter_actor_queue_redraw (self);
 }
 
 static void
@@ -362,6 +365,7 @@ st_scroll_view_set_property (GObject      *object,
     case PROP_AUTO_SCROLL:
       st_scroll_view_set_auto_scrolling (self,
                                          g_value_get_boolean (value));
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
@@ -403,6 +407,13 @@ st_scroll_view_dispose (GObject *object)
       priv->vadjustment = NULL;
     }
 
+  if (priv->setting_connect_id > 0) {
+    g_signal_handler_disconnect (priv->settings, priv->setting_connect_id);
+    priv->setting_connect_id = 0;
+  }
+
+  g_clear_object (&priv->settings);
+
   g_signal_handlers_disconnect_by_func (ST_SCROLL_VIEW (object), motion_event_cb, ST_SCROLL_VIEW (object));
 
   G_OBJECT_CLASS (st_scroll_view_parent_class)->dispose (object);
@@ -413,13 +424,13 @@ st_scroll_view_paint (ClutterActor *actor)
 {
   StScrollViewPrivate *priv = ST_SCROLL_VIEW (actor)->priv;
 
-  /* StBin will paint the child */
-  CLUTTER_ACTOR_CLASS (st_scroll_view_parent_class)->paint (actor);
+  st_widget_paint_background (ST_WIDGET (actor));
 
-  /* paint our custom children */
-  if (priv->hscrollbar_visible && CLUTTER_ACTOR_IS_VISIBLE (priv->hscroll))
+  if (priv->child)
+    clutter_actor_paint (priv->child);
+  if (priv->hscrollbar_visible)
     clutter_actor_paint (priv->hscroll);
-  if (priv->vscrollbar_visible && CLUTTER_ACTOR_IS_VISIBLE (priv->vscroll))
+  if (priv->vscrollbar_visible)
     clutter_actor_paint (priv->vscroll);
 }
 
@@ -432,10 +443,11 @@ st_scroll_view_pick (ClutterActor       *actor,
   /* Chain up so we get a bounding box pained (if we are reactive) */
   CLUTTER_ACTOR_CLASS (st_scroll_view_parent_class)->pick (actor, color);
 
-  /* paint our custom children */
-  if (priv->hscrollbar_visible && CLUTTER_ACTOR_IS_VISIBLE (priv->hscroll))
+  if (priv->child)
+    clutter_actor_paint (priv->child);
+  if (priv->hscrollbar_visible)
     clutter_actor_paint (priv->hscroll);
-  if (priv->vscrollbar_visible && CLUTTER_ACTOR_IS_VISIBLE (priv->vscroll))
+  if (priv->vscrollbar_visible)
     clutter_actor_paint (priv->vscroll);
 }
 
@@ -445,7 +457,7 @@ get_scrollbar_width (StScrollView *scroll,
 {
   StScrollViewPrivate *priv = scroll->priv;
 
-  if (CLUTTER_ACTOR_IS_VISIBLE (priv->vscroll))
+  if (clutter_actor_is_visible (priv->vscroll))
     {
       gfloat min_size;
 
@@ -463,7 +475,7 @@ get_scrollbar_height (StScrollView *scroll,
 {
   StScrollViewPrivate *priv = scroll->priv;
 
-  if (CLUTTER_ACTOR_IS_VISIBLE (priv->hscroll))
+  if (clutter_actor_is_visible (priv->hscroll))
     {
       gfloat min_size;
 
@@ -483,14 +495,14 @@ st_scroll_view_get_preferred_width (ClutterActor *actor,
                                     gfloat       *natural_width_p)
 {
   StScrollViewPrivate *priv = ST_SCROLL_VIEW (actor)->priv;
-  StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (actor));
-  gboolean account_for_vscrollbar = FALSE;
+  StThemeNode *theme_node;
   gfloat min_width = 0, natural_width;
   gfloat child_min_width, child_natural_width;
 
   if (!priv->child)
     return;
 
+  theme_node = st_widget_get_theme_node (ST_WIDGET (actor));
   st_theme_node_adjust_for_height (theme_node, &for_height);
 
   clutter_actor_get_preferred_width (priv->child, -1,
@@ -505,31 +517,17 @@ st_scroll_view_get_preferred_width (ClutterActor *actor,
       break;
     case GTK_POLICY_ALWAYS:
     case GTK_POLICY_AUTOMATIC:
+    case GTK_POLICY_EXTERNAL:
       /* Should theoretically use the min width of the hscrollbar,
        * but that's not cleanly defined at the moment */
       min_width = 0;
       break;
-    }
-
-  switch (priv->vscrollbar_policy)
-    {
-    case GTK_POLICY_NEVER:
-      account_for_vscrollbar = FALSE;
-      break;
-    case GTK_POLICY_ALWAYS:
-      account_for_vscrollbar = TRUE;
-      break;
-    case GTK_POLICY_AUTOMATIC:
-      /* For automatic scrollbars, we always request space for the vertical
-       * scrollbar; we won't know whether we actually need one until our
-       * height is assigned in allocate().
-       */
-      account_for_vscrollbar = TRUE;
+    default:
+      g_warn_if_reached();
       break;
     }
 
-  account_for_vscrollbar = priv->vscrollbar_policy != GTK_POLICY_NEVER;
-  if (account_for_vscrollbar)
+  if (priv->vscrollbar_policy != GTK_POLICY_NEVER)
     {
       float sb_width = get_scrollbar_width (ST_SCROLL_VIEW (actor), for_height);
 
@@ -553,7 +551,7 @@ st_scroll_view_get_preferred_height (ClutterActor *actor,
                                      gfloat       *natural_height_p)
 {
   StScrollViewPrivate *priv = ST_SCROLL_VIEW (actor)->priv;
-  StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (actor));
+  StThemeNode *theme_node = NULL;
   gboolean account_for_hscrollbar = FALSE;
   gfloat min_height = 0, natural_height;
   gfloat child_min_height, child_natural_height;
@@ -563,6 +561,8 @@ st_scroll_view_get_preferred_height (ClutterActor *actor,
   if (!priv->child)
     return;
 
+  theme_node = st_widget_get_theme_node (ST_WIDGET (actor));
+
   st_theme_node_adjust_for_width (theme_node, &for_width);
 
   clutter_actor_get_preferred_width (priv->child, -1,
@@ -576,17 +576,22 @@ st_scroll_view_get_preferred_height (ClutterActor *actor,
   switch (priv->vscrollbar_policy)
     {
     case GTK_POLICY_NEVER:
+    case GTK_POLICY_EXTERNAL:
       break;
     case GTK_POLICY_ALWAYS:
     case GTK_POLICY_AUTOMATIC:
       /* We've requested space for the scrollbar, subtract it back out */
       for_width -= sb_width;
       break;
+    default:
+      g_warn_if_reached();
+      break;
     }
 
   switch (priv->hscrollbar_policy)
     {
     case GTK_POLICY_NEVER:
+    case GTK_POLICY_EXTERNAL:
       account_for_hscrollbar = FALSE;
       break;
     case GTK_POLICY_ALWAYS:
@@ -595,6 +600,9 @@ st_scroll_view_get_preferred_height (ClutterActor *actor,
     case GTK_POLICY_AUTOMATIC:
       account_for_hscrollbar = for_width < child_min_width;
       break;
+    default:
+      g_warn_if_reached();
+      break;
     }
 
   clutter_actor_get_preferred_height (priv->child, for_width,
@@ -609,10 +617,14 @@ st_scroll_view_get_preferred_height (ClutterActor *actor,
       break;
     case GTK_POLICY_ALWAYS:
     case GTK_POLICY_AUTOMATIC:
+    case GTK_POLICY_EXTERNAL:
       /* Should theoretically use the min height of the vscrollbar,
        * but that's not cleanly defined at the moment */
       min_height = 0;
       break;
+    default:
+      g_warn_if_reached();
+      break;
     }
 
   if (account_for_hscrollbar)
@@ -638,24 +650,13 @@ st_scroll_view_allocate (ClutterActor          *actor,
                          ClutterAllocationFlags flags)
 {
   ClutterActorBox content_box, child_box;
-  ClutterActorClass *parent_parent_class;
   gfloat avail_width, avail_height, sb_width, sb_height;
   gboolean hscrollbar_visible, vscrollbar_visible;
 
   StScrollViewPrivate *priv = ST_SCROLL_VIEW (actor)->priv;
   StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (actor));
 
-  /* Chain up to the parent's parent class
-   *
-   * We do this because we do not want StBin to allocate the child, as we
-   * give it a different allocation later, depending on whether the scrollbars
-   * are visible
-   */
-  parent_parent_class
-    = g_type_class_peek_parent (st_scroll_view_parent_class);
-
-  CLUTTER_ACTOR_CLASS (parent_parent_class)->
-  allocate (actor, box, flags);
+  clutter_actor_set_allocation (actor, box, flags);
 
   st_theme_node_get_content_box (theme_node, box, &content_box);
 
@@ -738,7 +739,7 @@ st_scroll_view_allocate (ClutterActor          *actor,
    */
 
   /* Vertical scrollbar */
-  if (CLUTTER_ACTOR_IS_VISIBLE (priv->vscroll))
+  if (clutter_actor_is_visible (priv->vscroll))
     {
       if (st_widget_get_direction (ST_WIDGET (actor)) == ST_TEXT_DIRECTION_RTL)
         {
@@ -757,7 +758,7 @@ st_scroll_view_allocate (ClutterActor          *actor,
     }
 
   /* Horizontal scrollbar */
-  if (CLUTTER_ACTOR_IS_VISIBLE (priv->hscroll))
+  if (clutter_actor_is_visible (priv->hscroll))
     {
       if (st_widget_get_direction (ST_WIDGET (actor)) == ST_TEXT_DIRECTION_RTL)
         {
@@ -826,6 +827,12 @@ st_scroll_view_style_changed (StWidget *widget)
 
   StThemeNode *theme_node = st_widget_get_theme_node (widget);
   gdouble fade_offset = st_theme_node_get_length (theme_node, "-st-vfade-offset");
+
+  gboolean vfade_enabled = g_settings_get_boolean (priv->settings, "enable-vfade");
+
+  if (!vfade_enabled)
+    fade_offset = 0;
+
   st_scroll_view_update_vfade_effect (self, fade_offset);
 
   st_widget_style_changed (ST_WIDGET (priv->hscroll));
@@ -834,6 +841,15 @@ st_scroll_view_style_changed (StWidget *widget)
   ST_WIDGET_CLASS (st_scroll_view_parent_class)->style_changed (widget);
 }
 
+static void
+vfade_setting_changed_cb (GSettings *settings, gchar *key, gpointer user_data)
+{
+    StWidget *widget = ST_WIDGET (user_data);
+    g_return_if_fail (ST_IS_SCROLL_VIEW (widget));
+
+    st_scroll_view_style_changed (widget);
+}
+
 static gboolean
 st_scroll_view_scroll_event (ClutterActor       *self,
                              ClutterScrollEvent *event)
@@ -871,6 +887,9 @@ st_scroll_view_scroll_event (ClutterActor       *self,
                     "value", &value,
                     NULL);
       break;
+    default:
+      g_warn_if_reached();
+      break;
     }
 
   switch (event->direction)
@@ -891,8 +910,13 @@ st_scroll_view_scroll_event (ClutterActor       *self,
     case CLUTTER_SCROLL_RIGHT:
       st_adjustment_set_value (priv->hadjustment, value + step);
       break;
+    default:
+      g_warn_if_reached();
+      break;
     }
 
+  clutter_actor_queue_redraw (self);
+
   return TRUE;
 }
 
@@ -904,8 +928,6 @@ st_scroll_view_class_init (StScrollViewClass *klass)
   ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
   StWidgetClass *widget_class = ST_WIDGET_CLASS (klass);
 
-  g_type_class_add_private (klass, sizeof (StScrollViewPrivate));
-
   object_class->get_property = st_scroll_view_get_property;
   object_class->set_property = st_scroll_view_set_property;
   object_class->dispose = st_scroll_view_dispose;
@@ -989,7 +1011,7 @@ st_scroll_view_class_init (StScrollViewClass *klass)
 static void
 st_scroll_view_init (StScrollView *self)
 {
-  StScrollViewPrivate *priv = self->priv = SCROLL_VIEW_PRIVATE (self);
+  StScrollViewPrivate *priv = self->priv = st_scroll_view_get_instance_private (self);
 
   priv->hscrollbar_policy = GTK_POLICY_AUTOMATIC;
   priv->vscrollbar_policy = GTK_POLICY_AUTOMATIC;
@@ -1006,8 +1028,8 @@ st_scroll_view_init (StScrollView *self)
                                 "vertical", TRUE,
                                 NULL);
 
-  clutter_actor_set_parent (priv->hscroll, CLUTTER_ACTOR (self));
-  clutter_actor_set_parent (priv->vscroll, CLUTTER_ACTOR (self));
+  clutter_actor_add_child (CLUTTER_ACTOR (self), priv->hscroll);
+  clutter_actor_add_child (CLUTTER_ACTOR (self), priv->vscroll);
 
   /* mouse scroll is enabled by default, so we also need to be reactive */
   priv->mouse_scroll = TRUE;
@@ -1017,6 +1039,8 @@ st_scroll_view_init (StScrollView *self)
   priv->auto_scroll_timeout_id = 0;
   priv->mouse_pointer = clutter_device_manager_get_core_device (clutter_device_manager_get_default (),
                                                                       CLUTTER_POINTER_DEVICE);
+  priv->settings = g_settings_new("org.cinnamon");
+  priv->setting_connect_id = g_signal_connect (priv->settings, "changed::enable-vfade", G_CALLBACK (vfade_setting_changed_cb), self);
 }
 
 static void
@@ -1074,7 +1098,7 @@ st_scroll_view_remove (ClutterContainer *container,
       else
         g_assert ("Unknown child removed from StScrollView");
 
-      clutter_actor_unparent (actor);
+      clutter_actor_remove_child (CLUTTER_ACTOR (container), actor);
     }
 }
 
@@ -1272,6 +1296,10 @@ st_scroll_view_set_auto_scrolling (StScrollView *scroll,
                           G_CALLBACK (motion_event_cb), scroll);
       } else {
         g_signal_handlers_disconnect_by_func (scroll, motion_event_cb, scroll);
+        if (priv->auto_scroll_timeout_id > 0) {
+            g_source_remove (priv->auto_scroll_timeout_id);
+            priv->auto_scroll_timeout_id = 0;
+        }
       }
     }
 }
